from typing import List, Callable

from movado.controller import Controller
from movado.estimator import Estimator
from functools import wraps
from movado.distance_controller import DistanceController
from movado.hoeffding_adaptive_tree_model import HoeffdingAdaptiveTreeModel
from movado.chained_estimator import ChainedEstimator

# This imports are unused but populate the global symbol table for the globals() call
from movado.mab_controller import MabController
from movado.kernel_regression_model import KernelRegressionModel


def approximate(
    **kwargs,
) -> Callable[
    [Callable[[List[float]], List[float]]], Callable[[List[float]], List[float]]
]:
    def approximate_decorator(
        func: Callable[[List[float]], List[float]]
    ) -> Callable[[List[float]], List[float]]:
        controller: Controller
        estimator: Estimator
        is_first_call: bool = True

        @wraps(func)
        def wrapper(*wrapper_args) -> List[float]:
            # args must either contain the point to be analyzed
            # or it must contain self (for class methods) and then the point
            nonlocal is_first_call
            nonlocal controller
            nonlocal estimator
            if len(wrapper_args) < 1 or len(wrapper_args) > 2:
                raise Exception(
                    "The decorated function must have a single input which is a list of numbers, "
                    + "self may also be present for non-static class methods, in such case self is "
                    + "expected to be the first argument followed by the point and nothing"
                )
            if len(wrapper_args) == 1:
                point = wrapper_args[0]
            else:
                point = wrapper_args[1]
            if is_first_call:
                is_first_call = False
                selected_model = globals().get(str(kwargs.get("estimator")) + "Model")
                model = (
                    HoeffdingAdaptiveTreeModel if not selected_model else selected_model
                )
                # TODO add multi-output non-chained estimators
                outputs: int = kwargs.get("outputs")
                if not outputs:
                    raise Exception(
                        "Please specify outputs as a kwarg as the number of targets of the fitness function"
                    )
                estimator = ChainedEstimator(
                    model(
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "estimator"
                        }
                    ),
                    outputs,
                )
                selected_controller = globals().get(
                    str(kwargs.get("controller")) + "Controller"
                )
                controller = (
                    DistanceController
                    if not selected_controller
                    else selected_controller
                )
                controller = controller(
                    func,
                    estimator,
                    self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
                    **{
                        k[k.find("_") + 1 :]: v
                        for k, v in kwargs.items()
                        if k[: k.find("_")] == "controller"
                    },
                )

            return list(
                controller.compute_objective(point)
            )  # we expect the first and only argument of the function to be the input point

        if kwargs.get("disabled"):
            return func
        return wrapper

    return approximate_decorator


class Movado:
    def __init__(self):
        self.models = [
            model.replace("Model", "")
            for model in globals().keys()
            if ("Model" in model) and len(model) > 5
        ]
        self.controllers = [
            controller.replace("Controller", "")
            for controller in globals().keys()
            if ("Controller" in controller) and len(controller) > 10
        ]



from abc import abstractmethod, ABC
from typing import Tuple, Optional, Callable, Any, List, Dict, Union

from scipy.stats import PearsonRConstantInputWarning
from sklearn.preprocessing import StandardScaler
import numpy as np
import scipy as sp
import scipy.stats
import time
from pathlib import Path

from movado.estimator import Estimator
from movado.mab_handler import MabHandler
import warnings


class Controller(ABC):
    @abstractmethod
    def __init__(
        self,
        exact_fitness: Callable[[List[float]], List[float]],
        estimator: Estimator,
        self_exact: Optional[object] = None,
        debug=False,
        **kwargs
    ):
        self.__scaler = StandardScaler()
        self.__time_dataset: List[float] = []
        self.__error_dataset: List[float] = []
        self.__cost_dataset: List[float] = []
        self.__to_reshape = False
        self._exact: Callable[[List[float]], List[float]] = exact_fitness
        self._estimator: Estimator = estimator
        self._debug = debug
        self._self_exact = self_exact
        if self._debug:
            Path("movado_debug").mkdir(exist_ok=True)
            self._controller_debug = "movado_debug/controller.csv"
            Path(self._controller_debug).open("w").close()

    @abstractmethod
    def compute_objective(self, point: List[int]) -> List[float]:
        pass

    def _compute_exact(
        self,
        point: List[int],
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_forced_probability: Optional[int] = None,
        mab_weight: Optional[MabHandler] = None,
        mab_weight_forced_probability: Optional[int] = None,
    ) -> Tuple[List[float], float]:
        if self._self_exact:
            exact, exec_time = Controller.measure_execution_time(
                self._exact, self._self_exact, point
            )
        else:
            exact, exec_time = Controller.measure_execution_time(self._exact, point)

        self._estimator.train(point, exact)
        if mab:
            if mab_weight:
                time_weight = mab_weight.predict([mab[0].get_mean_cost()]) / 100
                mab[0].learn(
                    mab[1],
                    self.get_time_error_z_score(
                        exec_time, 0, *(time_weight, 1 - time_weight)
                    ),
                    point,
                    mab_forced_probability,
                ),
                mab_weight.learn(
                    time_weight * 100,
                    self.get_time_error_correlation(),
                    [mab[0].get_mean_cost()],
                    mab_weight_forced_probability,
                )
            else:
                mab[0].learn(
                    mab[1],
                    self.get_time_error_z_score(
                        exec_time,
                        0,
                    ),
                    point,
                    mab_forced_probability,
                )
        return exact, exec_time

    def _compute_estimated(
        self,
        point: List[int],
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_weight: MabHandler = None,
    ) -> Tuple[List[float], float]:
        estimation, exec_time = Controller.measure_execution_time(
            self._estimator.predict,
            point,
        )
        if mab:
            if mab_weight:
                time_weight = mab_weight.predict([mab[0].get_mean_cost()]) / 100
                mab[0].learn(
                    mab[1],
                    self.get_time_error_z_score(
                        exec_time,
                        self._estimator.get_error(),
                        *(time_weight, 1 - time_weight)
                    ),
                    point,
                )
                mab_weight.learn(
                    time_weight,
                    self.get_time_error_correlation(),
                    [mab[0].get_mean_cost()],
                )
            else:
                mab[0].learn(
                    mab[1],
                    self.get_time_error_z_score(
                        exec_time,
                        self._estimator.get_error(),
                    ),
                    point,
                )

        return estimation, exec_time

    @abstractmethod
    def _write_debug(self, debug_info: Dict[str, Any]):
        pass

    def get_time_error_z_score(
        self,
        response_time: float,
        error: float,
        time_weight: float = 1,
        error_weight: float = 1,
    ) -> float:

        self.__time_dataset.append(response_time)
        self.__error_dataset.append(error)
        if len(self.__time_dataset) > 1:
            reshaped_time = np.reshape(self.__time_dataset, (-1, 1))
            reshaped_error = np.reshape(self.__error_dataset, (-1, 1))
        else:
            reshaped_time = np.reshape(self.__time_dataset, (1, -1))
            reshaped_error = np.reshape(self.__error_dataset, (1, -1))
        self.__scaler.fit(reshaped_time)
        time_z_score: float = self.__scaler.transform(reshaped_time)[-1][0]
        self.__scaler.fit(reshaped_error)
        error_z_score: float = self.__scaler.transform(reshaped_error)[-1][0]
        out = (time_weight * time_z_score) + (error_weight * error_z_score)
        self.__cost_dataset.append(out)
        return out

    @staticmethod
    def __is_constant(x):
        return x.count(x[0]) == len(x)

    def get_time_error_correlation(self):
        try:
            samples = len(self.__error_dataset)
            independent_vars = 2
            if (
                self.__is_constant(self.__error_dataset)
                or self.__is_constant(self.__time_dataset)
                or self.__is_constant(self.__cost_dataset)
                or samples - independent_vars == 1
            ):
                return 0
            time_cost_correlation: float = sp.stats.pearsonr(
                self.__time_dataset, self.__cost_dataset
            )[0]
            error_cost_correlation: float = sp.stats.pearsonr(
                self.__error_dataset, self.__cost_dataset
            )[0]
            time_error_correlation: float = sp.stats.pearsonr(
                self.__time_dataset, self.__error_dataset
            )[0]
            R2 = np.sqrt(
                (
                    np.abs(time_cost_correlation ** 2)
                    + np.abs(error_cost_correlation ** 2)
                    - 2
                    * time_cost_correlation
                    * error_cost_correlation
                    * time_error_correlation
                )
                / (1 - np.abs(time_error_correlation))
            )
            R2_adj = 1 - (
                ((1 - R2 ** 2) * (samples - 1)) / (samples - independent_vars - 1)
            )
            return -R2_adj
        except ValueError:
            return 0

    @staticmethod
    def measure_execution_time(
        func: Callable[..., Any], *args, **kwargs
    ) -> Tuple[Any, float]:
        start = time.time()
        out = func(*args, **kwargs)
        execution_time = time.time() - start
        return out, execution_time



from pathlib import Path
from typing import List, Callable, Dict, Any, Optional

from movado.controller import Controller
from movado.estimator import Estimator
from movado.mab_handler_cb import MabHandlerCB
from movado.mab_handler_cats import MabHandlerCATS


class MabController(Controller):
    def __init__(
        self,
        exact_fitness: Callable[[List[float]], List[float]],
        estimator: Estimator,
        self_exact: Optional[object] = None,
        debug: bool = False,
        epsilon: float = 0.2,
    ):
        super().__init__(
            exact_fitness=exact_fitness,
            estimator=estimator,
            self_exact=self_exact,
            debug=debug,
        )
        self.__mab = MabHandlerCB(debug, epsilon=epsilon)
        self.__weight_mab = MabHandlerCATS(
            debug=debug, epsilon=epsilon, debug_path="mab_weight"
        )
        self.__is_first_call = True

        if self._debug:
            Path(self._controller_debug).open("a").write(
                "Point, Exec_Time, Error, Estimation\n"
            )

    def compute_objective(self, point: List[int]) -> List[float]:
        decision = self.__mab.predict([*point, self._estimator.get_error()])
        accuracy = self._estimator.get_error()
        if decision == 2 or accuracy == 0.0:
            out, exec_time = self._compute_exact(
                point,
                (self.__mab, 2),
                1 if accuracy == 0.0 else None,
                self.__weight_mab,
                1 if accuracy == 0.0 else None,
            )
        else:
            out, exec_time = self._compute_estimated(
                point, (self.__mab, 1), self.__weight_mab
            )

        # TODO probably this check can be done only once
        if self._debug:
            self._write_debug(
                {
                    "Point": point,
                    "Exec_Time": exec_time,
                    "Error": self._estimator.get_error(),
                    "Estimation": 0 if decision == 2 or accuracy == 0.0 else 1,
                }
            )
        self.__is_first_call = False
        return out

    def _write_debug(self, debug_info: Dict[str, Any]):
        Path(self._controller_debug).open("a").write(
            str(debug_info["Point"])
            + ", "
            + str(debug_info["Exec_Time"])
            + ", "
            + str(debug_info["Error"])
            + ", "
            + str(debug_info["Estimation"])
            + "\n"
        )

    def get_mean_cost(self):
        return self.__mab.get_mean_cost()



from pathlib import Path
from typing import Union, Optional, Dict, List, Any
import numpy as np
from sklearn.neighbors import KDTree, DistanceMetric

from movado.controller import Controller
from movado.mab_handler_cats import MabHandlerCATS


class DistanceController(Controller):
    # do not annotate types in __init__, it throws "ABCMeta Object is not subscriptable"
    def __init__(
        self,
        exact_fitness,
        estimator,
        self_exact: Optional[object] = None,
        nth_nearest=3,
        distance_metric="minkowski",
        threshold="mab",
        debug=False,
        **kwargs,
    ):
        super().__init__(
            exact_fitness=exact_fitness,
            estimator=estimator,
            self_exact=self_exact,
            debug=debug,
        )

        self.__evaluated_points: List[List[float]] = []
        self.__threshold: str = threshold
        self.__distance_metric = distance_metric
        self.__nth_nearest: int = nth_nearest
        self.__are_neighbours_to_recompute: bool = True
        self.__knn: Optional[KDTree] = None
        self.__distance_metric_kwargs: Dict[str, Union[int, float]] = {
            k: kwargs[k] for k in kwargs.keys() if k.startswith("distance_")
        }
        self.__threshold_kwargs: Dict[str, Union[int, float]] = {
            k: kwargs[k] for k in kwargs.keys() if k.startswith("threshold_")
        }
        self.__distances: List[float] = []
        if self.__threshold == "fixed" and not self.__threshold_kwargs.get(
            "threshold_value"
        ):
            raise Exception(
                "Missing keyword argument threshold_value, this argument is mandatory"
                + " when using fixed threshold "
            )
        self.__mab = None
        self.__weight_mab = None
        if self.__threshold == "mab":
            if not kwargs.get("mab_actions"):
                kwargs["mab_actions"] = 100
            if not kwargs.get("mab_bandwidth"):
                kwargs["mab_bandwidth"] = 5
            if kwargs.get("mab_epsilon"):
                self.__mab = MabHandlerCATS(
                    debug=debug, epsilon=kwargs.get("mab_epsilon")
                )
            else:
                self.__mab = MabHandlerCATS(debug=debug)
                self.__weight_mab = MabHandlerCATS(debug=debug, debug_path="mab_weight")
            if kwargs.get("mab_weight"):
                if kwargs.get("mab_epsilon"):
                    self.__weight_mab = MabHandlerCATS(
                        debug=debug,
                        epsilon=kwargs.get("mab_epsilon"),
                        debug_path="mab_weight",
                    )
                else:
                    self.__weight_mab = MabHandlerCATS(
                        debug=debug, debug_path="mab_weight"
                    )

        if self._debug:
            Path(self._controller_debug).open("a").write(
                "Threshold, Nth_Nearest_Distance, Point, Exec_Time, Error, Estimation\n"
            )

    def compute_objective(self, point: List[int]) -> List[float]:
        out: List[float]
        mae: float
        self.__add_point(point)
        threshold = self.__get_threshold(point)
        nth_distance = self.__get_nth_nearest_distance(point)
        error = self._estimator.get_error()

        if nth_distance > threshold or error == 0.0:
            out, exec_time = self._compute_exact(
                point,
                (self.__mab, threshold * 100) if self.__threshold == "mab" else None,
                1 if self.__threshold == "mab" and error == 0.0 else None,
                self.__weight_mab
                if self.__threshold == "mab" and self.__weight_mab
                else None,
                1
                if self.__threshold == "mab" and error == 0.0 and self.__weight_mab
                else None,
            )
        else:
            out, exec_time = self._compute_estimated(
                point,
                (self.__mab, threshold * 100) if self.__threshold == "mab" else None,
                self.__weight_mab
                if self.__threshold == "mab" and self.__weight_mab
                else None,
            )

        # TODO probably this check can be done only once
        if self._debug:
            self._write_debug(
                {
                    "Threshold": 0 if error == 0.0 else threshold,
                    "Nth_Nearest_Distance": 0
                    if not self.__evaluated_points
                    else nth_distance,
                    "Point": point,
                    "Exec_Time": exec_time,
                    "Error": error
                    if nth_distance > threshold or error == 0.0
                    else self._estimator.get_error(),
                    "Estimation": 0 if nth_distance > threshold or error == 0.0 else 1,
                }
            )
        return out

    def _write_debug(self, debug_info: Dict[str, Any]):
        Path(self._controller_debug).open("a").write(
            str(debug_info["Threshold"])
            + ", "
            + str(debug_info["Nth_Nearest_Distance"])
            + ", "
            + str(debug_info["Point"])
            + ", "
            + str(debug_info["Exec_Time"])
            + ", "
            + str(debug_info["Error"])
            + ", "
            + str(debug_info["Estimation"])
            + "\n"
        )

    def __add_point(self, point: List[float]) -> None:
        self.__evaluated_points.append(point)
        self.__are_neighbours_to_recompute = True

    def __get_threshold(self, point: List[int]) -> float:
        return getattr(self, "_get_threshold_" + self.__threshold)(
            point if self.__threshold == "mab" else None
        )

    def _get_threshold_fixed(self, *args) -> float:
        return self.__threshold_kwargs["threshold_value"]

    def _get_threshold_mean(self, *args):
        return np.mean(self._get_distances())

    def _get_threshold_median(self, *args):
        return np.median(self._get_distances())

    def _get_threshold_mab(self, point: List[int]) -> float:
        return self.__mab.predict(point) / 100

    def _get_distances(self) -> np.ndarray:
        distances: np.ndarray = np.empty([0])
        for point in self.__evaluated_points:
            distances = np.append(distances, self.__get_nth_nearest_distance(point))
        return distances

    def __get_nth_nearest_distance(self, point: List[float]) -> float:
        if self.__are_neighbours_to_recompute:
            self.__knn = KDTree(
                self.__evaluated_points,
                metric=DistanceMetric.get_metric(
                    self.__distance_metric, **self.__distance_metric_kwargs
                ),
            )
        k = min(self.__nth_nearest, len(self.__evaluated_points))
        distance = self.__knn.query([point], k=k)[0][0][k - 1]
        self.__distances.append(distance)
        diameter = max(self.__distances) - min(self.__distances)
        if diameter == 0:
            return 0
        else:
            return distance - min(self.__distances) / diameter
